"""

Overview
--------

A Scheduler is used to generate the order in which the nodes of a graph are
executed. By default, a Scheduler executes nodes in an order determined by the pattern of edges among the nodes in the graph, with each node executed once
per `PASS` through the graph. For example, in a ``graph`` in which a node *A* projects to a node
*B* that projects to a node *C*, *A* will execute first followed by *B*, and then *C* in each `PASS` through the
graph. However, a Scheduler can be used to implement more complex patterns of execution, by specifying
`Conditions <Condition>` that determine when and how many times individual nodes execute, and whether and how
this depends on the execution of other nodes. Any executable node in a graph can be assigned a
Condition, and Conditions can be combined in arbitrary ways to generate any pattern of execution of the nodes
in a graph that is logically possible.

.. _Scheduler_Creation:

Creating a Scheduler
--------------------

When creating a Scheduler explicitly, the set of nodes
to be executed and their order must be specified in the Scheduler's constructor using one the following:

* a *graph specification dictionary* in the **graph** argument -
  each entry of the dictionary must be a node of a graph, and the value of each entry must be a set of
  zero or more nodes that project directly to the key.  The graph must be acyclic; an error is generated if any
  cycles (e.g., recurrent dependencies) are detected.  The Scheduler computes a `toposort` from the graph that is
  used as the default order of executions, subject to any Conditions that have been specified
  (see `below <Scheduler_Algorithm>`).

Conditions can be added to a Scheduler when it is created by specifying a `ConditionSet` (a set of
`Conditions <Condition>`) in the **conditions** argument of its constructor.  Individual Conditions and/or
ConditionSets can also be added after the Scheduler has been created, using its `add_condition` and
`add_condition_set` methods, respectively.

.. _Scheduler_Algorithm:

Algorithm
---------

.. _Consideration_Set:

When a Scheduler is created, it constructs a `consideration_queue`:  a list of ``consideration sets``
that defines the order in which nodes are eligible to be executed.  This is based on the dependencies specified in the graph
specification provided in the Scheduler's constructor.  Each ``consideration_set``
is a set of nodes that are eligible to execute at the same time/`CONSIDERATION_SET_EXECUTION` (i.e.,
that appear at the same "depth" in a sequence of dependencies, and among which there are no dependencies).  The first
``consideration_set`` consists of only origin nodes. The second consists of all nodes
that receive edges from the nodes in the first ``consideration_set``.
The third consists of nodes that receive edges from nodes in the first two ``consideration sets``,
and so forth.  When the Scheduler is run, it uses the `consideration_queue` to determine which
nodes are eligible to execute in each `CONSIDERATION_SET_EXECUTION` of a `PASS`, and then evaluates the `Condition <Condition>`
associated with each node in the current ``consideration_set`` to determine which should
actually be assigned for execution.

Pseudocode::

    consideration_queue <- list(toposort(graph))

    reset TimeScale.ENVIRONMENT_STATE_UPDATE counters
    while TimeScale.ENVIRONMENT_STATE_UPDATE are not satisfied
    and TimeScale.ENVIRONMENT_SEQUENCE termination conditions are not satisfied:
        reset TimeScale.PASS counters
        cur_index <- 0

        while TimeScale.ENVIRONMENT_STATE_UPDATE termination conditions are not satisfied
        and TimeScale.ENVIRONMENT_SEQUENCE termination conditions are not satisfied
        and cur_index < len(consideration_queue):

            cur_consideration_set <- consideration_queue[cur_index]
            do:
                cur_consideration_set_has_changed <- False
                for cur_node in cur_consideration_set:
                    if  cur_node not in cur_consideration_set_execution
                        and cur_node`s Condition is satisfied:

                        cur_consideration_set_has_changed <- True
                        add cur_node to cur_consideration_set_execution
                        increment execution and time counters
            while cur_consideration_set_has_changed

            if cur_consideration_set_execution is not empty or absolute time conditions are used:
                yield cur_consideration_set_execution

            increment cur_index
            increment time counters

        if all execution sets yielded were empty:
            yield an empty execution set

.. _Scheduler_Execution:

Execution
---------

.. note::
    This section covers normal scheduler execution
    (`Scheduler.mode = SchedulingMode.STANDARD`). See
    `Scheduler_Exact_Time` below for a description of
    `exact time mode <SchedulingMode.EXACT_TIME>`.

When a Scheduler is run, it provides a set of nodes that should be run next, based on their dependencies in the
graph specification, and any `Conditions <Condition>`, specified in the Scheduler's
constructor. For each call to the `run <Scheduler.run>` method, the Scheduler sequentially evaluates its
`consideration sets <Consideration_Set>` in their order in the `consideration_queue`.  For each set, it determines
which nodes in the set are allowed to execute, based on whether their associated `Condition <Condition>` has
been met. Any node that does not have a `Condition` explicitly specified is assigned a Condition that causes it
to be executed whenever it is `under consideration <Scheduler_Algorithm>` and all its structural parents have been
executed at least once since the node's last execution. All of the nodes within a `consideration_set
<consideration_set>` that are allowed to execute comprise a `CONSIDERATION_SET_EXECUTION` of execution. These nodes are considered
as executing simultaneously.

.. note::
    The ordering of the nodes specified within a `CONSIDERATION_SET_EXECUTION` is arbitrary (and is irrelevant, as there are no
    graph dependencies among nodes within the same ``consideration_set``). However,
    the execution of a node within a `CONSIDERATION_SET_EXECUTION` may trigger the execution of another node within its
    ``consideration_set``, as in the example below::

           C
          ↗ ↖
         A   B

        scheduler.add_condition(B, graph_scheduler.EveryNCalls(A, 2))
        scheduler.add_condition(C, graph_scheduler.EveryNCalls(B, 1))

        execution sets: [{A}, {A, B}, {C}, ...]

    Since there are no graph dependencies between `A` and `B`, they may execute in the same `CONSIDERATION_SET_EXECUTION`. Morever,
    `A` and `B` are in the same ``consideration_set``. Since `B` is specified to run every two
    times `A` runs, `A`'s second execution in the second `CONSIDERATION_SET_EXECUTION` allows `B` to run within that `CONSIDERATION_SET_EXECUTION`,
    rather than waiting for the next `PASS`.

For each `CONSIDERATION_SET_EXECUTION`, the Scheduler evaluates whether any specified
`termination Conditions <Scheduler_Termination_Conditions>` have been met, and terminates if so.  Otherwise,
it returns the set of nodes that should be executed in the current `CONSIDERATION_SET_EXECUTION`. Each subsequent call to the
`run <Scheduler.run>` method returns the set of nodes in the following `CONSIDERATION_SET_EXECUTION`.

Processing of all of the `consideration_sets <consideration_set>` in the `consideration_queue` constitutes a `PASS` of
execution, over which every node in the graph has been considered for execution. Subsequent calls to the
`run <Scheduler.run>` method cycle back through the `consideration_queue`, evaluating the `consideration_sets
<consideration_set>` in the same order as previously. Different subsets of nodes within the same `consideration_set
<consideration_set>` may be assigned to execute on each `PASS`, since different Conditions may be satisfied.

The Scheduler continues to make `PASS`\\ es through the `consideration_queue` until a `termination Condition
<Scheduler_Termination_Conditions>` is satisfied. If no termination Conditions are specified, by default the Scheduler
terminates an `ENVIRONMENT_STATE_UPDATE <TimeScale.ENVIRONMENT_STATE_UPDATE>` when every node has been specified for execution at least once
(corresponding to the `AllHaveRun` Condition).  However, other termination Conditions can be specified,
that may cause the Scheduler to terminate an `ENVIRONMENT_STATE_UPDATE <TimeScale.ENVIRONMENT_STATE_UPDATE>` earlier or later (e.g., when the Condition
for a particular node or set of nodes is met).

.. _Scheduler_Termination_Conditions:

*Termination Conditions*
~~~~~~~~~~~~~~~~~~~~~~~~

Termination conditions are `Conditions <Condition>` that specify when the open-ended units of time - `ENVIRONMENT_STATE_UPDATE
<TimeScale.ENVIRONMENT_STATE_UPDATE>` and `ENVIRONMENT_SEQUENCE` - have ended.  By default, the termination condition for an `ENVIRONMENT_STATE_UPDATE <TimeScale.ENVIRONMENT_STATE_UPDATE>` is
`AllHaveRun`, which is satisfied when all nodes have run at least once within the environment state update, and the termination
condition for an `ENVIRONMENT_SEQUENCE` is when all of its constituent environment state updates have terminated.


.. _Scheduler_Absolute_Time:

Absolute Time
-------------

The scheduler supports scheduling of models of real-time systems in
modes, both of which involve mapping real-time values to
`Time`. The default mode is is most compatible with standard
scheduling, but can cause some unexpected behavior in certain cases
because it is inexact. The consideration queue remains intact, but as a
result, actions specified by fixed times of absolute-time-based
conditions (`start <TimeInterval.start>` and `end <TimeInterval.end>` of
`TimeInterval`, and `t` of `TimeTermination`) may not occur at exactly
the time specified. The simplest example of this situation involves a
linear graph with two nodes::

    >>> import graph_scheduler

    >>> graph = {'A': set(), 'B': {'A'}}
    >>> scheduler = graph_scheduler.Scheduler(graph=graph)

    >>> scheduler.add_condition('A', graph_scheduler.TimeInterval(start=10))
    >>> scheduler.add_condition('B', graph_scheduler.TimeInterval(start=10))

In standard mode, **A** and **B** are in different consideration sets,
and so can never execute at the same time. At most one of **A** and
**B** will start exactly at t=10ms, with the other starting at its next
consideration after. There are many of these examples, and while it may
be solveable in some cases, it is not a simple problem. So,
`Exact Time Mode <Scheduler_Exact_Time>` exists as an alternative
option for these cases, though it comes with its own drawbacks.

.. note::
    Due to issues with floating-point precision, absolute time values in
    conditions and `Time` are limited to 8 decimal points. If more
    precision is needed, use
    `fractions <https://docs.python.org/3/library/fractions.html>`_,
    where possible, or smaller units (e.g. microseconds instead of
    milliseconds).


.. _Scheduler_Exact_Time:

Exact Time Mode
~~~~~~~~~~~~~~~

When `Scheduler.mode` is `SchedulingMode.EXACT_TIME`, the scheduler is
capable of handling examples like the one
`above <Scheduler_Absolute_Time>`. In this mode, all nodes in the
scheduler's graph become members of the same consideration set, and may
be executed at the same time for every consideration set execution,
subject to the conditions specified. As a result, the guarantees in
`standard scheduling <Scheduler_Execution>` may not apply - that is,
that all parent nodes get a chance to execute before their children, and
that there exist no data dependencies (edges) between nodes in the
same execution set. In exact time mode, all nodes will be in one
[unordered] execution set. An ordering may be inferred by the original
graph, however, using the `indices in the original consideration queue\
<graph_scheduler.scheduler.Scheduler.consideration_queue_indices>`.
Additionally, non-absolute conditions like
`EveryNCalls` may behave unexpectedly in some cases.

Examples
--------

Please see `Condition` for a list of all supported Conditions and their behavior.

* Basic phasing in a linear process::

    >>> import graph_scheduler

    >>> graph = {'A': set(), 'B': {'A'}, 'C': {'B'}}
    >>> scheduler = graph_scheduler.Scheduler(graph=graph)

    >>> # implicit condition of Always for A
    >>> scheduler.add_condition('B', graph_scheduler.EveryNCalls('A', 2))
    >>> scheduler.add_condition('C', graph_scheduler.EveryNCalls('B', 3))

    >>> # implicit AllHaveRun Termination condition
    >>> execution_sequence = list(scheduler.run())
    >>> execution_sequence
    [{'A'}, {'A'}, {'B'}, {'A'}, {'A'}, {'B'}, {'A'}, {'A'}, {'B'}, {'C'}]

* Alternate basic phasing in a linear process::

    >>> graph = {'A': set(), 'B': {'A'}}
    >>> scheduler = graph_scheduler.Scheduler(graph=graph)

    >>> scheduler.add_condition(
    ...     'A',
    ...     graph_scheduler.Any(
    ...         graph_scheduler.AtPass(0),
    ...         graph_scheduler.EveryNCalls('B', 2)
    ...     )
    ... )

    >>> scheduler.add_condition(
    ...     'B',
    ...     graph_scheduler.Any(
    ...         graph_scheduler.EveryNCalls('A', 1),
    ...         graph_scheduler.EveryNCalls('B', 1)
    ...     )
    ... )
    >>> termination_conds = {
    ...     graph_scheduler.TimeScale.ENVIRONMENT_STATE_UPDATE: graph_scheduler.AfterNCalls('B', 4, time_scale=graph_scheduler.TimeScale.ENVIRONMENT_STATE_UPDATE)
    ... }
    >>> execution_sequence = list(scheduler.run(termination_conds=termination_conds))
    >>> execution_sequence
    [{'A'}, {'B'}, {'B'}, {'A'}, {'B'}, {'B'}]

* Basic phasing in two processes::

    >>> graph = {'A': set(), 'B': set(), 'C': {'A', 'B'}}
    >>> scheduler = graph_scheduler.Scheduler(graph=graph)

    >>> scheduler.add_condition('A', graph_scheduler.EveryNPasses(1))
    >>> scheduler.add_condition('B', graph_scheduler.EveryNCalls('A', 2))
    >>> scheduler.add_condition(
    ...     'C',
    ...     graph_scheduler.Any(
    ...         graph_scheduler.AfterNCalls('A', 3),
    ...         graph_scheduler.AfterNCalls('B', 3)
    ...     )
    ... )
    >>> termination_conds = {
    ...     graph_scheduler.TimeScale.ENVIRONMENT_STATE_UPDATE: graph_scheduler.AfterNCalls('C', 4, time_scale=graph_scheduler.TimeScale.ENVIRONMENT_STATE_UPDATE)
    ... }
    >>> execution_sequence = list(scheduler.run(termination_conds=termination_conds))
    >>> execution_sequence # doctest: +SKIP
    [{'A'}, {'A', 'B'}, {'A'}, {'C'}, {'A', 'B'}, {'C'}, {'A'}, {'C'}, {'A', 'B'}, {'C'}]

.. _Scheduler_Class_Reference

Class Reference
===============

"""

import copy
import datetime
import enum
import fractions
import logging
import typing

import numpy as np
import pint
from toposort import toposort

from graph_scheduler import _unit_registry
from graph_scheduler.condition import (
    All, AllHaveRun, Always, Condition, ConditionSet, EveryNCalls, Never,
    _parse_absolute_unit, _quantity_as_integer,
)
from graph_scheduler.time import Clock, TimeScale

__all__ = [
    'Scheduler', 'SchedulerError', 'SchedulingMode',
]

logger = logging.getLogger(__name__)

default_termination_conds = {
    TimeScale.ENVIRONMENT_SEQUENCE: Never(),
    TimeScale.ENVIRONMENT_STATE_UPDATE: AllHaveRun(),
}


class SchedulingMode(enum.Enum):
    """
        Attributes:
            STANDARD
                `Standard Mode <Scheduler_Execution>`

            EXACT_TIME
                `Exact time Mode <Scheduler_Exact_Time>`
    """
    STANDARD = enum.auto()
    EXACT_TIME = enum.auto()


class SchedulerError(Exception):

    def __init__(self, error_value):
        self.error_value = error_value

    def __str__(self):
        return repr(self.error_value)


class Scheduler:
    """Generates an order of execution for nodes in a graph or graph
    specification dictionary, possibly determined by a set of `Conditions <Condition>`.

    Arguments
    ---------

    graph : Dict[object: set(object)], `networkx.DiGraph`
        a graph specification dictionary - each entry of the dictionary must be an object,
        and the value of each entry must be a set of zero or more objects that project directly to the key.

    conditions  : ConditionSet
        set of `Conditions <Condition>` that specify when individual nodes in **graph**
        execute and any dependencies among them.

    mode : SchedulingMode[STANDARD|EXACT_TIME] : SchedulingMode.STANDARD
        sets the mode of scheduling: `standard <Scheduler_Execution>` or
        `exact time <Scheduler_Exact_Time>`

    default_absolute_time_unit : `pint.Quantity` : ``1ms``
        if not otherwise determined by any absolute **conditions**,
        specifies the absolute duration of a `CONSIDERATION_SET_EXECUTION`

    Attributes
    ----------

    conditions : ConditionSet
        the set of Conditions the Scheduler uses when running

    default_execution_id : object
        the execution_id to use if none is supplied to `run`; a unique
        identifier to allow multiple schedulings independently. Must be
        hashable.

    execution_list : list
        the full history of consideration set executions the Scheduler has produced

    consideration_queue : list
        a list form of the Scheduler's toposort ordering of its nodes

    consideration_queue_indices : dict
        a dict mapping **nodes** to their position in the `consideration_queue`

    termination_conds : Dict[TimeScale: Condition]
        a mapping from `TimeScales <TimeScale>` to `Conditions <Condition>` that, when met, terminate the execution
        of the specified `TimeScale`. On set, update only for the
        `TimeScale`\\ s specified in the argument.

    mode
        sets the mode of scheduling: `standard <Scheduler_Execution>` or
        `exact time <Scheduler_Exact_Time>`

        :type: SchedulingMode
        :default: `SchedulingMode.STANDARD`

    default_absolute_time_unit
        if not otherwise determined by any absolute **conditions**,
        specifies the absolute duration of a `CONSIDERATION_SET_EXECUTION`

        :type: `pint.Quantity`
        :default: ``1ms``

    """
    def __init__(
        self,
        graph,
        conditions=None,
        termination_conds=None,
        default_execution_id=None,
        mode: SchedulingMode = SchedulingMode.STANDARD,
        default_absolute_time_unit: typing.Union[str, pint.Quantity] = 1 * _unit_registry.ms,
        **kwargs
    ):
        """
        :param self:
        :param conditions: (ConditionSet) - a :keyword:`ConditionSet` to be scheduled
        """
        self.conditions = ConditionSet(conditions)

        # the consideration queue is the ordered list of sets of nodes in the graph, by the
        # order in which they should be checked to ensure that all parents have a chance to run before their children
        self.consideration_queue = []
        if termination_conds is None:
            termination_conds = default_termination_conds.copy()
        else:
            termination_conds = {**default_termination_conds, **termination_conds}
        self.default_termination_conds = Scheduler._parse_termination_conditions(termination_conds)
        self._termination_conds = self.default_termination_conds.copy()

        self.cycle_nodes = set()
        self.mode = mode
        self.default_absolute_time_unit = _parse_absolute_unit(default_absolute_time_unit)

        if graph is not None:
            try:
                # networkx graph
                self.dependency_dict = {
                    child: set(parents.keys())
                    for child, parents in graph.succ.items()
                }
            except AttributeError:
                self.dependency_dict = graph
            self.consideration_queue = list(toposort(self.dependency_dict))
            self.nodes = []
            for consideration_set in self.consideration_queue:
                for node in consideration_set:
                    self.nodes.append(node)
        else:
            raise SchedulerError(
                'Must instantiate a Scheduler with a graph dependency dict or a networkx.DiGraph'
            )

        self._generate_consideration_queue_indices()

        self.default_execution_id = default_execution_id
        # stores the in order list of self.run's yielded outputs
        self.execution_list = {self.default_execution_id: []}
        self.execution_timestamps = {self.default_execution_id: []}
        self.clocks = {self.default_execution_id: Clock()}
        self.counts_total = {}
        self.counts_useable = {}
        self._init_counts(execution_id=self.default_execution_id)
        self.date_creation = datetime.datetime.now()
        self.date_last_run_end = None

    def _generate_consideration_queue_indices(self):
        self.consideration_queue_indices = {}
        for i, cs in enumerate(self.consideration_queue):
            self.consideration_queue_indices.update({
                n: i for n in cs
            })

    def _init_counts(self, execution_id, base_execution_id=NotImplemented):
        """
            Attributes
            ----------

                execution_id
                    the execution_id to initialize counts for

                base_execution_id
                    if specified, the counts for execution_id will be copied from the counts of base_execution_id
                    default : NotImplemented
        """
        # all counts are divided by execution_id, which provides a context for the scheduler's execution, so that
        # it can be reused in multiple contexts

        # stores total the number of occurrences of a node through the time scale
        # i.e. the number of times node has ran/been queued to run in a environment state update
        if execution_id not in self.counts_total:
            if base_execution_id is not NotImplemented:
                if base_execution_id not in self.counts_total:
                    raise SchedulerError('execution_id {0} not in {1}.counts_total'.format(base_execution_id, self))

                self.counts_total[execution_id] = {
                    ts: {n: self.counts_total[base_execution_id][ts][n] for n in self.nodes} for ts in TimeScale
                }
            else:
                self.counts_total[execution_id] = {
                    ts: {n: 0 for n in self.nodes} for ts in TimeScale
                }

        # counts_useable is a dictionary intended to store the number of available "instances" of a certain node that
        # are available to expend in order to satisfy conditions such as "run B every two times A runs"
        # specifically, counts_useable[a][b] = n indicates that there are n uses of a that are available for b to expend
        # so, in the previous example B would check to see if counts_useable[A][B] >= 2, in which case B can run
        # then, counts_useable[a][b] would be reset to 0, even if it was greater than 2
        if execution_id not in self.counts_useable:
            if base_execution_id is not NotImplemented:
                if base_execution_id not in self.counts_useable:
                    raise SchedulerError('execution_id {0} not in {1}.counts_useable'.format(base_execution_id, self))

                self.counts_useable[execution_id] = {
                    node: {n: self.counts_useable[base_execution_id][node][n] for n in self.nodes} for node in self.nodes
                }
            else:
                self.counts_useable[execution_id] = {
                    node: {n: 0 for n in self.nodes} for node in self.nodes
                }

        if execution_id not in self.execution_list:
            if base_execution_id is not NotImplemented:
                if base_execution_id not in self.execution_list:
                    raise SchedulerError('execution_id {0} not in {1}.execution_list'.format(base_execution_id, self))

                self.execution_list[execution_id] = list(self.execution_list[base_execution_id])
            else:
                self.execution_list[execution_id] = []

        self._init_clock(execution_id, base_execution_id)

    def _init_clock(self, execution_id, base_execution_id=NotImplemented):
        # instantiate new Clock for this execution_id if necessary
        # currently does not work with base_execution_id
        if execution_id not in self.clocks:
            if base_execution_id is not NotImplemented:
                if base_execution_id not in self.clocks:
                    raise SchedulerError('execution_id {0} not in {1}.clocks'.format(base_execution_id, self))

                self.clocks[execution_id] = copy.deepcopy(self.clocks[base_execution_id])
            else:
                self.clocks[execution_id] = Clock()

    def _delete_counts(self, execution_id):
        for obj in [self.counts_useable, self.counts_total, self.clocks, self.execution_list]:
            try:
                del obj[execution_id]
            except KeyError:
                pass

    def _reset_counts_total(self, time_scale, execution_id):
        for ts in TimeScale:
            # only reset the values underneath the current scope
            # this works because the enum is set so that higher granularities of time have lower values
            if ts.value <= time_scale.value:
                for c in self.counts_total[execution_id][ts]:
                    self.counts_total[execution_id][ts][c] = 0

    def _reset_counts_useable(self, execution_id):
        self.counts_useable[execution_id] = {
            node: {n: 0 for n in self.nodes} for node in self.nodes
        }

    def _combine_termination_conditions(self, termination_conds):
        termination_conds = Scheduler._parse_termination_conditions(termination_conds)
        new_conds = self.termination_conds.copy()
        new_conds.update(termination_conds)

        return new_conds

    @staticmethod
    def _parse_termination_conditions(termination_conds):
        # parse string representation of TimeScale
        parsed_conds = {}
        delkeys = set()
        for scale in termination_conds:
            try:
                parsed_conds[getattr(TimeScale, scale.upper())] = termination_conds[scale]
                delkeys.add(scale)
            except (AttributeError, TypeError):
                pass

        termination_conds.update(parsed_conds)

        try:
            termination_conds = {
                k: termination_conds[k] for k in termination_conds
                if (
                    isinstance(k, TimeScale)
                    and isinstance(termination_conds[k], Condition)
                    and k not in delkeys
                )
            }
        except TypeError:
            raise TypeError('termination_conditions must be a dictionary of the form {TimeScale: Condition, ...}')
        else:
            return termination_conds

    ################################################################################
    # Wrapper methods
    #   to allow the user to ignore the ConditionSet internals
    ################################################################################
    def __contains__(self, item):
        return self.conditions.__contains__(item)

    def add_condition(self, owner, condition):
        """
        Adds a `Condition` to the Scheduler. If **owner** already has a Condition, it is overwritten
        with the new one. If you want to add multiple conditions to a single owner, use the
        `composite Conditions <Conditions_Composite>` to accurately specify the desired behavior.

        Arguments
        ---------

        owner : ``node``
            specifies the node with which the **condition** should be associated. **condition**
            will govern the execution behavior of **owner**

        condition : Condition
            specifies the Condition, associated with the **owner** to be added to the ConditionSet.
        """
        self.conditions.add_condition(owner, condition)

    def add_condition_set(self, conditions):
        """
        Adds a set of `Conditions <Condition>` (in the form of a dict or another ConditionSet) to the Scheduler.
        Any Condition added here will overwrite an existing Condition for a given owner.
        If you want to add multiple conditions to a single owner, add a single `Composite Condition <Conditions_Composite>`
        to accurately specify the desired behavior.

        Arguments
        ---------

        conditions : dict[``node``: `Condition`], `ConditionSet`
            specifies collection of Conditions to be added to this ConditionSet,

            if a dict is provided:
                each entry should map an owner node (the node whose execution behavior will be
                governed) to a `Condition <Condition>`

        """
        self.conditions.add_condition_set(conditions)

    ################################################################################
    # Validation methods
    #   to provide the user with info if they do something odd
    ################################################################################
    def _validate_run_state(self):
        self._validate_conditions()

    def _validate_conditions(self):
        unspecified_nodes = []
        for node in self.nodes:
            if node not in self.conditions:
                dependencies = list(self.dependency_dict[node])
                if len(dependencies) == 0:
                    cond = Always()
                elif len(dependencies) == 1:
                    cond = EveryNCalls(dependencies[0], 1)
                else:
                    cond = All(*[EveryNCalls(x, 1) for x in dependencies])

                self.add_condition(node, cond)
                unspecified_nodes.append(node)
        if len(unspecified_nodes) > 0:
            logger.info(
                'These nodes have no Conditions specified, and will be scheduled with conditions: {0}'.format(
                    {node: self.conditions[node] for node in unspecified_nodes}
                )
            )

    ################################################################################
    # Run methods
    ################################################################################
    def run(
        self,
        termination_conds=None,
        execution_id=None,
        base_execution_id=None,
        skip_environment_state_update_time_increment=False,
        **kwargs,
    ):
        """
        run is a python generator, that when iterated over provides the next `CONSIDERATION_SET_EXECUTION` of
        executions at each iteration

        :param termination_conds: (dict) - a mapping from `TimeScale`\\s to `Condition`\\s that when met
               terminate the execution of the specified `TimeScale`
        """
        self._validate_run_state()

        if self.mode is SchedulingMode.EXACT_TIME:
            effective_consideration_queue = [set(self.nodes)]
        else:
            effective_consideration_queue = self.consideration_queue

        if termination_conds is None:
            termination_conds = self.termination_conds
        else:
            termination_conds = self._combine_termination_conditions(termination_conds)

        current_time = self.get_clock(execution_id).time
        is_satisfied_kwargs = {
            'scheduler': self,
            'execution_id': execution_id,
            **kwargs
        }

        in_absolute_time_mode = len(self.get_absolute_conditions(termination_conds)) > 0
        if in_absolute_time_mode:
            current_time.absolute_interval = self._get_absolute_consideration_set_execution_unit(termination_conds)
            # advance absolute clock time to first necessary time
            current_time.absolute = max(
                current_time.absolute,
                min([
                    min(c.absolute_fixed_points)
                    if len(c.absolute_fixed_points) > 0 else 0
                    for c in self.get_absolute_conditions(termination_conds).values()
                ])
            )
            # convert to interval time unit to avoid pint floating point issues
            current_time.absolute = current_time.absolute.to(current_time.absolute_interval.u)

            # .to auto converts to float even if magnitude is an int, undo
            try:
                current_time.absolute = _quantity_as_integer(current_time.absolute)
            except ValueError:
                pass

        current_time.absolute_enabled = in_absolute_time_mode

        self._init_counts(execution_id, base_execution_id)
        self._reset_counts_useable(execution_id)
        self._reset_counts_total(TimeScale.ENVIRONMENT_STATE_UPDATE, execution_id)

        while (
            not termination_conds[TimeScale.ENVIRONMENT_STATE_UPDATE].is_satisfied(**is_satisfied_kwargs)
            and not termination_conds[TimeScale.ENVIRONMENT_SEQUENCE].is_satisfied(**is_satisfied_kwargs)
        ):
            self._reset_counts_total(TimeScale.PASS, execution_id)

            execution_list_has_changed = False
            cur_index_consideration_queue = 0

            while (
                cur_index_consideration_queue < len(effective_consideration_queue)
                and not termination_conds[TimeScale.ENVIRONMENT_STATE_UPDATE].is_satisfied(**is_satisfied_kwargs)
                and not termination_conds[TimeScale.ENVIRONMENT_SEQUENCE].is_satisfied(**is_satisfied_kwargs)
            ):
                # all nodes to be added during this consideration set execution
                cur_consideration_set_execution_exec = set()
                # the current "layer/group" of nodes that MIGHT be added during this consideration set execution
                cur_consideration_set = effective_consideration_queue[cur_index_consideration_queue]

                try:
                    iter(cur_consideration_set)
                except TypeError as e:
                    raise SchedulerError('cur_consideration_set is not iterable, did you ensure that this Scheduler was instantiated with an actual toposort output for param toposort_ordering? err: {0}'.format(e))

                # do-while, on cur_consideration_set_has_changed
                # we check whether each node in the current consideration set is allowed to run,
                # and nodes can cause cascading adds within this set
                while True:
                    cur_consideration_set_has_changed = False
                    for current_node in cur_consideration_set:
                        # only add each node once during a single consideration set execution, this also serves
                        # to prevent infinitely cascading adds
                        if current_node not in cur_consideration_set_execution_exec:
                            if self.conditions.conditions[current_node].is_satisfied(**is_satisfied_kwargs):
                                cur_consideration_set_execution_exec.add(current_node)
                                execution_list_has_changed = True
                                cur_consideration_set_has_changed = True

                                for ts in TimeScale:
                                    self.counts_total[execution_id][ts][current_node] += 1
                                # current_node's node is added to the execution queue, so we now need to
                                # reset all of the counts useable by current_node's node to 0
                                for n in self.counts_useable[execution_id]:
                                    self.counts_useable[execution_id][n][current_node] = 0
                                # and increment all of the counts of current_node's node useable by other
                                # nodes by 1
                                for n in self.counts_useable[execution_id]:
                                    self.counts_useable[execution_id][current_node][n] += 1
                    # do-while condition
                    if not cur_consideration_set_has_changed:
                        break

                # add a new consideration set execution at each step in a pass, if the consideration set execution would not be empty
                if len(cur_consideration_set_execution_exec) >= 1 or in_absolute_time_mode:
                    self.execution_list[execution_id].append(cur_consideration_set_execution_exec)
                    if in_absolute_time_mode:
                        self.execution_timestamps[execution_id].append(
                            copy.copy(current_time)
                        )
                    yield self.execution_list[execution_id][-1]

                    self.get_clock(execution_id)._increment_time(TimeScale.CONSIDERATION_SET_EXECUTION)

                cur_index_consideration_queue += 1

            # if an entire pass occurs with nothing running, add an empty consideration set execution
            if not execution_list_has_changed and not in_absolute_time_mode:
                self.execution_list[execution_id].append(set())
                yield self.execution_list[execution_id][-1]

                self.get_clock(execution_id)._increment_time(TimeScale.CONSIDERATION_SET_EXECUTION)

            self.get_clock(execution_id)._increment_time(TimeScale.PASS)

        if not skip_environment_state_update_time_increment:
            self.get_clock(execution_id)._increment_time(TimeScale.ENVIRONMENT_STATE_UPDATE)

        if termination_conds[TimeScale.ENVIRONMENT_SEQUENCE].is_satisfied(**is_satisfied_kwargs):
            self.date_last_run_end = datetime.datetime.now()

        return self.execution_list[execution_id]

    def _increment_time(self, time_scale, execution_id):
        self.get_clock(execution_id)._increment_time(time_scale)

    def _get_absolute_consideration_set_execution_unit(self, termination_conds=None):
        """Computes the time length of the gap between two consideration set executions
        """
        if termination_conds is None:
            termination_conds = self.termination_conds

        # all of the units of time that must occur on a consideration set execution
        intervals = []
        for c in self.get_absolute_conditions(termination_conds).values():
            intervals.extend(c.absolute_intervals)
            if self.mode is SchedulingMode.EXACT_TIME:
                intervals.extend(c.absolute_fixed_points)

        if len(intervals) == 0:
            return self.default_absolute_time_unit
        else:
            min_time_unit = min(intervals, key=lambda x: x.u).u

            # convert all intervals into the same unit
            for i, a in enumerate(intervals):
                unit_conversion = round(np.log10((1 * a.u).to(min_time_unit).m))
                intervals[i] = (int(a.m) * 10 ** unit_conversion) * min_time_unit

            # numerator is the largest possible length of a pass
            # denominator evenly divides this length by number of consideration set executions
            numerator = np.gcd.reduce([interval.m for interval in intervals])
            if self.mode == SchedulingMode.STANDARD:
                denominator = len(self.consideration_queue)
            elif self.mode == SchedulingMode.EXACT_TIME:
                denominator = 1

            if denominator == 1:
                res = numerator
            else:
                res = fractions.Fraction(numerator=numerator, denominator=denominator)
            return res * min_time_unit

    def get_absolute_conditions(self, termination_conds=None):
        if termination_conds is None:
            termination_conds = self.termination_conds

        return {
            owner: cond
            for owner, cond
            in [*self.conditions.conditions.items(), *termination_conds.items()]
            if cond.is_absolute
        }

    def get_clock(self, execution_id):
        try:
            return self.clocks[execution_id.default_execution_id]
        except AttributeError:
            try:
                return self.clocks[execution_id]
            except AttributeError:
                if execution_id not in self.clocks:
                    self._init_clock(execution_id)

                return self.clocks[execution_id]
        except KeyError:
            raise

    @property
    def termination_conds(self):
        return self._termination_conds

    @termination_conds.setter
    def termination_conds(self, termination_conds):
        """Updates this Scheduler's base `termination conditions
        <Scheduler.termination_conds>`_ to be used on future `run
        <Scheduler.run>`\\ s for which termination conditions are not
        specified.

        Arguments:
            termination_conds : dict[[TimeScale, str]: Condition]
                the dictionary of termination Conditions to overwrite
                the current base
                `termination conditions <Scheduler.termination_conds>`
        """
        if termination_conds is None:
            self._termination_conds = self.default_termination_conds.copy()
        else:
            self._termination_conds = self._combine_termination_conditions(
                termination_conds
            )

    @property
    def _in_exact_time_mode(self):
        return self.mode is SchedulingMode.EXACT_TIME or len(self.consideration_queue) == 1



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

from sphinx.ext.autodoc import between

import graph_scheduler._version

sys.path.insert(0, os.path.abspath('../src/'))


# -- Project information -----------------------------------------------------

project = 'Graph Scheduler'
copyright = '2021, Princeton University'
author = 'Katherine Mantel'


release = graph_scheduler._version.get_versions()['version']
version = '.'.join(release.split('.')[:3])


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.intersphinx',
    'sphinx.ext.napoleon',
    'sphinx_autodoc_typehints',
    'sphinx_rtd_theme',
    'm2r2',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

autodoc_member_order = 'bysource'
default_role = 'any'

intersphinx_mapping = {
    'networkx': ('https://networkx.org/documentation/stable/', None),
    'pint': ('https://pint.readthedocs.io/en/stable/', None),
}


def setup(app):
    # Register a sphinx.ext.autodoc.between listener to ignore everything
    # between lines that contain the word COMMENT
    app.connect('autodoc-process-docstring', between('^.*COMMENT.*$', exclude=True))
    return app



import inspect
import logging
import weakref

logger = logging.getLogger(__name__)

_unused_args_sig_cache = weakref.WeakKeyDictionary()


def prune_unused_args(func, args=None, kwargs=None):
    """
        Arguments
        ---------
            func : function

            args : *args

            kwargs : **kwargs


        Returns
        -------
            a tuple such that the first item is the intersection of **args** and the
            positional arguments of **func**, and the second item is the intersection
            of **kwargs** and the keyword arguments of **func**

    """
    # use the func signature to filter out arguments that aren't compatible
    try:
        sig = _unused_args_sig_cache[func]
    except KeyError:
        sig = inspect.signature(func)
        _unused_args_sig_cache[func] = sig

    has_args_param = False
    has_kwargs_param = False
    count_positional = 0
    func_kwargs_names = set()

    for name, param in sig.parameters.items():
        if param.kind is inspect.Parameter.VAR_POSITIONAL:
            has_args_param = True
        elif param.kind is inspect.Parameter.VAR_KEYWORD:
            has_kwargs_param = True
        elif param.kind is inspect.Parameter.POSITIONAL_OR_KEYWORD or param.kind is inspect.Parameter.KEYWORD_ONLY:
            if param.default is inspect.Parameter.empty:
                count_positional += 1
            func_kwargs_names.add(name)

    if args is not None:
        try:
            args = list(args)
        except TypeError:
            args = [args]

        if not has_args_param:
            num_extra_args = len(args) - count_positional
            if num_extra_args > 0:
                logger.debug('{1} extra arguments specified to function {0}, will be ignored (values: {2})'.format(func, num_extra_args, args[-num_extra_args:]))
            args = args[:count_positional]
    else:
        args = []

    if kwargs is not None:
        kwargs = dict(kwargs)

        if not has_kwargs_param:
            filtered = set()
            for kw in kwargs:
                if kw not in func_kwargs_names:
                    filtered.add(kw)
            if len(filtered) > 0:
                logger.debug('{1} extra keyword arguments specified to function {0}, will be ignored (values: {2})'.format(func, len(filtered), filtered))
            for kw in filtered:
                del kwargs[kw]
    else:
        kwargs = {}

    return args, kwargs


def call_with_pruned_args(func, *args, **kwargs):
    """
        Calls **func** with only the **args** and **kwargs** that
        exist in its signature
    """
    args, kwargs = prune_unused_args(func, args, kwargs)
    return func(*args, **kwargs)



import psyneulink as pnl
import pytest

import graph_scheduler
from graph_scheduler.time import (
    SimpleTime, Time, TimeHistoryTree, TimeScale, _time_scale_aliases,
    remove_time_scale_alias, set_time_scale_alias,
)


class TestTime:
    @pytest.mark.parametrize(
        'base, increment_time_scale, expected',
        [
            (
                Time(environment_sequence=0, environment_state_update=0, pass_=0, consideration_set_execution=0),
                TimeScale.ENVIRONMENT_STATE_UPDATE,
                Time(environment_sequence=0, environment_state_update=1, pass_=0, consideration_set_execution=0),
            ),
            (
                Time(environment_sequence=0, environment_state_update=0, pass_=5, consideration_set_execution=9),
                TimeScale.ENVIRONMENT_STATE_UPDATE,
                Time(environment_sequence=0, environment_state_update=1, pass_=0, consideration_set_execution=0),
            ),
            (
                Time(environment_sequence=1, environment_state_update=0, pass_=5, consideration_set_execution=9),
                TimeScale.ENVIRONMENT_STATE_UPDATE,
                Time(environment_sequence=1, environment_state_update=1, pass_=0, consideration_set_execution=0),
            ),
            (
                Time(environment_sequence=1, environment_state_update=0, pass_=5, consideration_set_execution=9),
                TimeScale.CONSIDERATION_SET_EXECUTION,
                Time(environment_sequence=1, environment_state_update=0, pass_=5, consideration_set_execution=10),
            ),
        ],
    )
    def test_increment(self, base, increment_time_scale, expected):
        base._increment_by_time_scale(increment_time_scale)
        assert base == expected

    @pytest.mark.psyneulink
    def test_multiple_runs(self):
        t1 = pnl.TransferMechanism()
        t2 = pnl.TransferMechanism()

        C = pnl.Composition(pathways=[t1, t2])

        C.run(inputs={t1: [[1.0], [2.0], [3.0]]})
        assert C.scheduler.get_clock(C).time == pnl.Time(run=1, trial=0, pass_=0, time_step=0)

        C.run(inputs={t1: [[4.0], [5.0], [6.0]]})
        assert C.scheduler.get_clock(C).time == pnl.Time(run=2, trial=0, pass_=0, time_step=0)


class TestTimeHistoryTree:
    def test_defaults(self):
        h = TimeHistoryTree()

        for node in [h, h.children[0]]:
            assert len(node.children) == 1
            assert all([node.total_times[ts] == 0 for ts in node.total_times])
            assert node.time_scale == TimeScale.get_parent(node.children[0].time_scale)
            assert node.time_scale >= TimeScale.ENVIRONMENT_STATE_UPDATE

    @pytest.mark.parametrize(
        'max_depth',
        [
            (TimeScale.ENVIRONMENT_SEQUENCE),
            (TimeScale.ENVIRONMENT_STATE_UPDATE)
        ])
    def test_max_depth(self, max_depth):
        h = TimeHistoryTree(max_depth=max_depth)

        node = h
        found_max_depth = h.time_scale == max_depth
        while len(node.children) > 0:
            node = node.children[0]
            found_max_depth = found_max_depth or node.time_scale == max_depth
            assert node.time_scale >= max_depth

        assert found_max_depth


class TestAliasTimeScale:
    @pytest.fixture(scope='class', autouse=True)
    def setup_alias(cls):
        # must save and replace psyneulink aliases as long as tests still
        # depend on psyneulink
        existing_aliases = _time_scale_aliases.copy()
        for ts, alias in existing_aliases.items():
            remove_time_scale_alias(alias)

        set_time_scale_alias('MY_ENVIRONMENT_STATE_UPDATE_ALIAS', TimeScale.ENVIRONMENT_STATE_UPDATE)

        yield

        remove_time_scale_alias('MY_ENVIRONMENT_STATE_UPDATE_ALIAS')
        for ts, alias in existing_aliases.items():
            set_time_scale_alias(alias, ts)

    @staticmethod
    def assert_environment_state_update_and_alias_equals(time_obj, value):
        assert time_obj.my_environment_state_update_alias == value
        assert time_obj.environment_state_update == value

        if isinstance(time_obj, Time):
            assert time_obj._get_by_time_scale(TimeScale.ENVIRONMENT_STATE_UPDATE) == value
            assert time_obj._get_by_time_scale(TimeScale.MY_ENVIRONMENT_STATE_UPDATE_ALIAS) == value

    def test_alias_references_timescale(self):
        assert TimeScale.MY_ENVIRONMENT_STATE_UPDATE_ALIAS is TimeScale.ENVIRONMENT_STATE_UPDATE
        assert TimeScale.get_parent(TimeScale.MY_ENVIRONMENT_STATE_UPDATE_ALIAS) is TimeScale.get_parent(TimeScale.ENVIRONMENT_STATE_UPDATE)
        assert TimeScale.get_child(TimeScale.MY_ENVIRONMENT_STATE_UPDATE_ALIAS) is TimeScale.get_child(TimeScale.ENVIRONMENT_STATE_UPDATE)

    @pytest.mark.parametrize(
        'kwargs, value',
        [
            ({'environment_state_update': 1}, 1),
            ({'my_environment_state_update_alias': 1}, 1)
        ]
    )
    def test_time_change_in_init(self, kwargs, value):
        t = Time(**kwargs)
        st = SimpleTime(t)

        self.assert_environment_state_update_and_alias_equals(t, value)
        self.assert_environment_state_update_and_alias_equals(st, value)

    def test_time_change_by_original_attr(self):
        t = Time()
        st = SimpleTime(t)

        t.environment_state_update = 1

        self.assert_environment_state_update_and_alias_equals(t, 1)
        self.assert_environment_state_update_and_alias_equals(st, 1)

    def test_time_change_by_alias_attr(self):
        t = Time()
        st = SimpleTime(t)

        t.my_environment_state_update_alias = 1

        self.assert_environment_state_update_and_alias_equals(t, 1)
        self.assert_environment_state_update_and_alias_equals(st, 1)

    @pytest.mark.parametrize('time_scale', [TimeScale.ENVIRONMENT_STATE_UPDATE, 'MY_ENVIRONMENT_STATE_UPDATE_ALIAS'])
    def test_time_change_by_method_set(self, time_scale):
        # alias does not exist at parametrization time
        try:
            time_scale = getattr(TimeScale, time_scale)
        except TypeError:
            pass

        t = Time()
        st = SimpleTime(t)
        t._set_by_time_scale(time_scale, 1)

        self.assert_environment_state_update_and_alias_equals(t, 1)
        self.assert_environment_state_update_and_alias_equals(st, 1)

    @pytest.mark.parametrize('time_scale', [TimeScale.ENVIRONMENT_STATE_UPDATE, 'MY_ENVIRONMENT_STATE_UPDATE_ALIAS'])
    def test_time_change_by_method_increment(self, time_scale):
        # alias does not exist at parametrization time
        try:
            time_scale = getattr(TimeScale, time_scale)
        except TypeError:
            pass

        t = Time()
        st = SimpleTime(t)
        t._increment_by_time_scale(time_scale)

        self.assert_environment_state_update_and_alias_equals(t, 1)
        self.assert_environment_state_update_and_alias_equals(st, 1)

    def test_time_repr(self):
        t = Time(my_environment_state_update_alias=1)
        st = SimpleTime(t)

        assert repr(t) == 'Time(environment_sequence: 0, my_environment_state_update_alias: 1, pass: 0, consideration_set_execution: 0)'
        assert repr(st) == 'Time(environment_sequence: 0, my_environment_state_update_alias: 1, consideration_set_execution: 0)'

    def test_aliased_conditions(self):
        graph = {'A': set()}
        sched_orig = graph_scheduler.Scheduler(
            graph=graph,
            termination_conds={TimeScale.ENVIRONMENT_STATE_UPDATE: graph_scheduler.AtPass(2)}
        )
        sched_alias = graph_scheduler.Scheduler(
            graph=graph,
            termination_conds={TimeScale.ENVIRONMENT_STATE_UPDATE: graph_scheduler.AtPass(2)}
        )

        sched_orig.add_condition('A', graph_scheduler.BeforeEnvironmentStateUpdate(3))
        sched_alias.add_condition('A', graph_scheduler.BeforeMyEnvironmentStateUpdateAlias(3))

        # should run in environment state updates 0-2 and not in 3
        for _ in range(4):
            assert list(sched_orig.run()) == list(sched_alias.run())



import doctest
import pytest

from psyneulink import clear_registry, primary_registries
from psyneulink.core import llvm as pnlvm


def pytest_runtest_setup(item):
    if 'cuda' in item.keywords and not pnlvm.ptx_enabled:
            pytest.skip('PTX engine not enabled/available')

    doctest.ELLIPSIS_MARKER = "[...]"

def pytest_generate_tests(metafunc):
    mech_and_func_modes = ['Python',
                           pytest.param('LLVM', marks=pytest.mark.llvm),
                           pytest.param('PTX', marks=[pytest.mark.llvm,
                                                      pytest.mark.cuda])
                          ]

    if "func_mode" in metafunc.fixturenames:
        metafunc.parametrize("func_mode", mech_and_func_modes)

    if "mech_mode" in metafunc.fixturenames:
        metafunc.parametrize("mech_mode", mech_and_func_modes)

    if "comp_mode_no_llvm" in metafunc.fixturenames:
        modes = [m for m in get_comp_execution_modes()
                 if m.values[0] is not pnlvm.ExecutionMode.LLVM]
        metafunc.parametrize("comp_mode", modes)

    elif "comp_mode" in metafunc.fixturenames:
        metafunc.parametrize("comp_mode", get_comp_execution_modes())

    if "autodiff_mode" in metafunc.fixturenames:
        auto_modes = [pnlvm.ExecutionMode.Python,
                      pytest.param(pnlvm.ExecutionMode.LLVMRun, marks=pytest.mark.llvm)]
        metafunc.parametrize("autodiff_mode", auto_modes)


def pytest_runtest_teardown(item):
    for registry in primary_registries:
        # Clear Registry to have a stable reference for indexed suffixes of default names
        clear_registry(registry)

    pnlvm.cleanup()

@pytest.fixture
def comp_mode_no_llvm():
    # dummy fixture to allow 'comp_mode' filtering
    pass

@pytest.helpers.register
def get_comp_execution_modes():
    return [pytest.param(pnlvm.ExecutionMode.Python),
            pytest.param(pnlvm.ExecutionMode.LLVM, marks=pytest.mark.llvm),
            pytest.param(pnlvm.ExecutionMode.LLVMExec, marks=pytest.mark.llvm),
            pytest.param(pnlvm.ExecutionMode.LLVMRun, marks=pytest.mark.llvm),
            pytest.param(pnlvm.ExecutionMode.PTXExec, marks=[pytest.mark.llvm, pytest.mark.cuda]),
            pytest.param(pnlvm.ExecutionMode.PTXRun, marks=[pytest.mark.llvm,  pytest.mark.cuda])
           ]


# TODO: remove this helper when tests no longer use psyneulink
@pytest.helpers.register
def composition_to_scheduler_args(composition):
    return {
        'graph': composition.graph_processing.prune_feedback_edges()[0],
        'default_execution_id': composition.default_execution_id
    }



import logging

import pytest
from psyneulink import _unit_registry
from psyneulink.core.components.functions.nonstateful.transferfunctions import Linear
from psyneulink.core.components.mechanisms.processing.transfermechanism import TransferMechanism
from psyneulink.core.components.projections.pathway.mappingprojection import MappingProjection
from psyneulink.core.compositions.composition import Composition
from psyneulink.core.scheduling.condition import (
    AfterCall, AfterNCalls, AfterNCallsCombined, AfterNPasses, AfterNEnvironmentStateUpdates,
    AfterPass, AfterEnvironmentStateUpdate, All, AllHaveRun, Always, Any, AtPass, AtConsiderationSetExecution,
    AtEnvironmentStateUpdate, AtEnvironmentStateUpdateStart, BeforeNCalls, BeforePass, BeforeConsiderationSetExecution,
    BeforeEnvironmentStateUpdate, Condition, ConditionError, EveryNCalls, EveryNPasses, Not,
    NWhen, TimeInterval, TimeTermination, WhenFinished, WhenFinishedAll,
    WhenFinishedAny, WhileNot,
)
from psyneulink.core.scheduling.scheduler import Scheduler
from psyneulink.core.scheduling.time import TimeScale

logger = logging.getLogger(__name__)


class TestCondition:

    def test_invalid_input_WhenFinished(self):
        with pytest.raises(ConditionError):
            WhenFinished(None).is_satisfied()

    def test_invalid_input_WhenFinishedAny_1(self):
        with pytest.raises(ConditionError):
            WhenFinished(None).is_satisfied()

    def test_invalid_input_WhenFinishedAny_2(self):
        with pytest.raises(ConditionError):
            WhenFinished({None}).is_satisfied()

    def test_invalid_input_WhenFinishedAll_1(self):
        with pytest.raises(ConditionError):
            WhenFinished(None).is_satisfied()

    def test_invalid_input_WhenFinishedAll_2(self):
        with pytest.raises(ConditionError):
            WhenFinished({None}).is_satisfied()

    def test_additional_args(self):
        class OneSatisfied(Condition):
            def __init__(self, a):
                def func(a, b):
                    return a or b
                super().__init__(func, a)

        cond = OneSatisfied(True)
        assert cond.is_satisfied(True)
        assert cond.is_satisfied(False)

        cond = OneSatisfied(False)
        assert cond.is_satisfied(True)
        assert not cond.is_satisfied(False)

    def test_additional_kwargs(self):
        class OneSatisfied(Condition):
            def __init__(self, a, c=True):
                def func(a, b, c=True):
                    return a or b or c
                super().__init__(func, a, c=True)

        cond = OneSatisfied(True)
        assert cond.is_satisfied(True)
        assert cond.is_satisfied(False, c=True)
        assert cond.is_satisfied(False, c=False)

        cond = OneSatisfied(True, c=False)
        assert cond.is_satisfied(True)
        assert cond.is_satisfied(False, c=True)
        assert cond.is_satisfied(False, c=False)

        cond = OneSatisfied(False)
        assert cond.is_satisfied(True)
        assert cond.is_satisfied(False, c=True)
        assert not cond.is_satisfied(False, c=False)
        assert not cond.is_satisfied(False, c=False, extra_arg=True)

    @pytest.mark.psyneulink
    class TestGeneric:
        def test_WhileNot_AtPass(self):
            comp = Composition()
            A = TransferMechanism(function=Linear(slope=5.0, intercept=2.0), name='A')
            comp.add_node(A)

            sched = Scheduler(**pytest.helpers.composition_to_scheduler_args(comp))
            sched.add_condition(A, WhileNot(lambda sched: sched.get_clock(sched.default_execution_id).get_total_times_relative(TimeScale.PASS, TimeScale.ENVIRONMENT_STATE_UPDATE) == 0, sched))

            termination_conds = {}
            termination_conds[TimeScale.ENVIRONMENT_SEQUENCE] = AfterNEnvironmentStateUpdates(1)
            termination_conds[TimeScale.ENVIRONMENT_STATE_UPDATE] = AtPass(5)
            output = list(sched.run(termination_conds=termination_conds))

            expected_output = [set(), A, A, A, A]
            assert output == pytest.helpers.setify_expected_output(expected_output)

        def test_WhileNot_AtPass_in_middle(self):
            comp = Composition()
            A = TransferMechanism(function=Linear(slope=5.0, intercept=2.0), name='A')
            comp.add_node(A)

            sched = Scheduler(**pytest.helpers.composition_to_scheduler_args(comp))
            sched.add_condition(A, WhileNot(lambda sched: sched.get_clock(sched.default_execution_id).get_total_times_relative(TimeScale.PASS, TimeScale.ENVIRONMENT_STATE_UPDATE) == 2, sched))

            termination_conds = {}
            termination_conds[TimeScale.ENVIRONMENT_SEQUENCE] = AfterNEnvironmentStateUpdates(1)
            termination_conds[TimeScale.ENVIRONMENT_STATE_UPDATE] = AtPass(5)
            output = list(sched.run(termination_conds=termination_conds))

            expected_output = [A, A, set(), A, A]
            assert output == pytest.helpers.setify_expected_output(expected_output)

    @pytest.mark.psyneulink
    class TestRelative:

        def test_Any_end_before_one_finished(self):
            comp = Composition()
            A = TransferMechanism(function=Linear(slope=5.0, intercept=2.0), name='A')
            for m in [A]:
                comp.add_node(m)
            sched = Scheduler(**pytest.helpers.composition_to_scheduler_args(comp))

            sched.add_condition(A, EveryNPasses(1))

            termination_conds = {}
            termination_conds[TimeScale.ENVIRONMENT_SEQUENCE] = AfterNEnvironmentStateUpdates(1)
            termination_conds[TimeScale.ENVIRONMENT_STATE_UPDATE] = Any(AfterNCalls(A, 10), AtPass(5))
            output = list(sched.run(termination_conds=termination_conds))

            expected_output = [A for _ in range(5)]
            assert output == pytest.helpers.setify_expected_output(expected_output)

        def test_All_end_after_one_finished(self):
            comp = Composition()
            A = TransferMechanism(function=Linear(slope=5.0, intercept=2.0), name='A')
            for m in [A]:
                comp.add_node(m)
            sched = Scheduler(**pytest.helpers.composition_to_scheduler_args(comp))

            sched.add_condition(A, EveryNPasses(1))

            termination_conds = {}
            termination_conds[TimeScale.ENVIRONMENT_SEQUENCE] = AfterNEnvironmentStateUpdates(1)
            termination_conds[TimeScale.ENVIRONMENT_STATE_UPDATE] = Any(AfterNCalls(A, 5), AtPass(10))
            output = list(sched.run(termination_conds=termination_conds))

            expected_output = [A for _ in range(5)]
            assert output == pytest.helpers.setify_expected_output(expected_output)

        def test_Not_AtPass(self):
            comp = Composition()
            A = TransferMechanism(function=Linear(slope=5.0, intercept=2.0), name='A')
            comp.add_node(A)

            sched = Scheduler(**pytest.helpers.composition_to_scheduler_args(comp))
            sched.add_condition(A, Not(AtPass(0)))

            termination_conds = {}
            termination_conds[TimeScale.ENVIRONMENT_SEQUENCE] = AfterNEnvironmentStateUpdates(1)
            termination_conds[TimeScale.ENVIRONMENT_STATE_UPDATE] = AtPass(5)
            output = list(sched.run(termination_conds=termination_conds))

            expected_output = [set(), A, A, A, A]
            assert output == pytest.helpers.setify_expected_output(expected_output)

        def test_Not_AtPass_in_middle(self):
            comp = Composition()
            A = TransferMechanism(function=Linear(slope=5.0, intercept=2.0), name='A')
            comp.add_node(A)

            sched = Scheduler(**pytest.helpers.composition_to_scheduler_args(comp))
            sched.add_condition(A, Not(AtPass(2)))

            termination_conds = {}
            termination_conds[TimeScale.ENVIRONMENT_SEQUENCE] = AfterNEnvironmentStateUpdates(1)
            termination_conds[TimeScale.ENVIRONMENT_STATE_UPDATE] = AtPass(5)
            output = list(sched.run(termination_conds=termination_conds))

            expected_output = [A, A, set(), A, A]
            assert output == pytest.helpers.setify_expected_output(expected_output)

        @pytest.mark.parametrize(
            'n,expected_output', [
                (0, ['A', 'A', 'A', 'A', 'A', 'A']),
                (1, ['A', 'A', 'A', 'B', 'A', 'A', 'A']),
                (2, ['A', 'A', 'A', 'B', 'A', 'B', 'A', 'A']),
            ]
        )
        def test_NWhen_AfterNCalls(self, n, expected_output):
            comp = Composition()
            A = TransferMechanism(function=Linear(slope=5.0, intercept=2.0), name='A')
            B = TransferMechanism(function=Linear(intercept=4.0), name='B')
            for m in [A, B]:
                comp.add_node(m)
            comp.add_projection(MappingProjection(), A, B)

            sched = Scheduler(**pytest.helpers.composition_to_scheduler_args(comp))
            sched.add_condition(A, Always())
            sched.add_condition(B, NWhen(AfterNCalls(A, 3), n))

            termination_conds = {}
            termination_conds[TimeScale.ENVIRONMENT_SEQUENCE] = AfterNEnvironmentStateUpdates(1)
            termination_conds[TimeScale.ENVIRONMENT_STATE_UPDATE] = AfterNCalls(A, 6)
            output = list(sched.run(termination_conds=termination_conds))

            expected_output = [A if x == 'A' else B for x in expected_output]

            assert output == pytest.helpers.setify_expected_output(expected_output)

    @pytest.mark.psyneulink
    class TestTime:

        def test_BeforeConsiderationSetExecution(self):
            comp = Composition()
            A = TransferMechanism(function=Linear(slope=5.0, intercept=2.0), name='A')
            comp.add_node(A)

            sched = Scheduler(**pytest.helpers.composition_to_scheduler_args(comp))
            sched.add_condition(A, BeforeConsiderationSetExecution(2))

            termination_conds = {}
            termination_conds[TimeScale.ENVIRONMENT_SEQUENCE] = AfterNEnvironmentStateUpdates(1)
            termination_conds[TimeScale.ENVIRONMENT_STATE_UPDATE] = AtPass(5)
            output = list(sched.run(termination_conds=termination_conds))

            expected_output = [A, A, set(), set(), set()]
            assert output == pytest.helpers.setify_expected_output(expected_output)

        def test_BeforeConsiderationSetExecution_2(self):
            comp = Composition()
            A = TransferMechanism(function=Linear(slope=5.0, intercept=2.0), name='A')
            B = TransferMechanism(name='B')
            comp.add_node(A)
            comp.add_node(B)

            comp.add_projection(MappingProjection(), A, B)

            sched = Scheduler(**pytest.helpers.composition_to_scheduler_args(comp))
            sched.add_condition(A, BeforeConsiderationSetExecution(2))
            sched.add_condition(B, Always())

            termination_conds = {}
            termination_conds[TimeScale.ENVIRONMENT_SEQUENCE] = AfterNEnvironmentStateUpdates(1)
            termination_conds[TimeScale.ENVIRONMENT_STATE_UPDATE] = AtPass(5)
            output = list(sched.run(termination_conds=termination_conds))

            expected_output = [A, B, B, B, B, B]
            assert output == pytest.helpers.setify_expected_output(expected_output)

        def test_AtConsiderationSetExecution(self):
            comp = Composition()
            A = TransferMechanism(function=Linear(slope=5.0, intercept=2.0), name='A')
            comp.add_node(A)

            sched = Scheduler(**pytest.helpers.composition_to_scheduler_args(comp))
            sched.add_condition(A, AtConsiderationSetExecution(0))

            termination_conds = {}
            termination_conds[TimeScale.ENVIRONMENT_SEQUENCE] = AfterNEnvironmentStateUpdates(1)
            termination_conds[TimeScale.ENVIRONMENT_STATE_UPDATE] = AtPass(5)
            output = list(sched.run(termination_conds=termination_conds))

            expected_output = [A, set(), set(), set(), set()]
            assert output == pytest.helpers.setify_expected_output(expected_output)

        def test_BeforePass(self):
            comp = Composition()
            A = TransferMechanism(function=Linear(slope=5.0, intercept=2.0), name='A')
            comp.add_node(A)

            sched = Scheduler(**pytest.helpers.composition_to_scheduler_args(comp))
            sched.add_condition(A, BeforePass(2))

            termination_conds = {}
            termination_conds[TimeScale.ENVIRONMENT_SEQUENCE] = AfterNEnvironmentStateUpdates(1)
            termination_conds[TimeScale.ENVIRONMENT_STATE_UPDATE] = AtPass(5)
            output = list(sched.run(termination_conds=termination_conds))

            expected_output = [A, A, set(), set(), set()]
            assert output == pytest.helpers.setify_expected_output(expected_output)

        def test_AtPass(self):
            comp = Composition()
            A = TransferMechanism(function=Linear(slope=5.0, intercept=2.0), name='A')
            comp.add_node(A)

            sched = Scheduler(**pytest.helpers.composition_to_scheduler_args(comp))
            sched.add_condition(A, AtPass(0))

            termination_conds = {}
            termination_conds[TimeScale.ENVIRONMENT_SEQUENCE] = AfterNEnvironmentStateUpdates(1)
            termination_conds[TimeScale.ENVIRONMENT_STATE_UPDATE] = AtPass(5)
            output = list(sched.run(termination_conds=termination_conds))

            expected_output = [A, set(), set(), set(), set()]
            assert output == pytest.helpers.setify_expected_output(expected_output)

        def test_AtPass_underconstrained(self):
            comp = Composition()
            A = TransferMechanism(function=Linear(slope=5.0, intercept=2.0), name='A')
            B = TransferMechanism(function=Linear(intercept=4.0), name='B')
            C = TransferMechanism(function=Linear(intercept=1.5), name='C')
            for m in [A, B, C]:
                comp.add_node(m)
            comp.add_projection(MappingProjection(), A, B)
            comp.add_projection(MappingProjection(), B, C)

            sched = Scheduler(**pytest.helpers.composition_to_scheduler_args(comp))
            sched.add_condition(A, AtPass(0))
            sched.add_condition(B, Always())
            sched.add_condition(C, Always())

            termination_conds = {}
            termination_conds[TimeScale.ENVIRONMENT_SEQUENCE] = AfterNEnvironmentStateUpdates(1)
            termination_conds[TimeScale.ENVIRONMENT_STATE_UPDATE] = AfterNCalls(C, 2)
            output = list(sched.run(termination_conds=termination_conds))

            expected_output = [A, B, C, B, C]
            assert output == pytest.helpers.setify_expected_output(expected_output)

        def test_AtPass_in_middle(self):
            comp = Composition()
            A = TransferMechanism(function=Linear(slope=5.0, intercept=2.0), name='A')
            comp.add_node(A)

            sched = Scheduler(**pytest.helpers.composition_to_scheduler_args(comp))
            sched.add_condition(A, AtPass(2))

            termination_conds = {}
            termination_conds[TimeScale.ENVIRONMENT_SEQUENCE] = AfterNEnvironmentStateUpdates(1)
            termination_conds[TimeScale.ENVIRONMENT_STATE_UPDATE] = AtPass(5)
            output = list(sched.run(termination_conds=termination_conds))

            expected_output = [set(), set(), A, set(), set()]
            assert output == pytest.helpers.setify_expected_output(expected_output)

        def test_AtPass_at_end(self):
            comp = Composition()
            A = TransferMechanism(function=Linear(slope=5.0, intercept=2.0), name='A')
            comp.add_node(A)

            sched = Scheduler(**pytest.helpers.composition_to_scheduler_args(comp))
            sched.add_condition(A, AtPass(5))

            termination_conds = {}
            termination_conds[TimeScale.ENVIRONMENT_SEQUENCE] = AfterNEnvironmentStateUpdates(1)
            termination_conds[TimeScale.ENVIRONMENT_STATE_UPDATE] = AtPass(5)
            output = list(sched.run(termination_conds=termination_conds))

            expected_output = [set(), set(), set(), set(), set()]
            assert output == pytest.helpers.setify_expected_output(expected_output)

        def test_AtPass_after_end(self):
            comp = Composition()
            A = TransferMechanism(function=Linear(slope=5.0, intercept=2.0), name='A')
            comp.add_node(A)

            sched = Scheduler(**pytest.helpers.composition_to_scheduler_args(comp))
            sched.add_condition(A, AtPass(6))

            termination_conds = {}
            termination_conds[TimeScale.ENVIRONMENT_SEQUENCE] = AfterNEnvironmentStateUpdates(1)
            termination_conds[TimeScale.ENVIRONMENT_STATE_UPDATE] = AtPass(5)
            output = list(sched.run(termination_conds=termination_conds))

            expected_output = [set(), set(), set(), set(), set()]
            assert output == pytest.helpers.setify_expected_output(expected_output)

        def test_AfterPass(self):
            comp = Composition()
            A = TransferMechanism(function=Linear(slope=5.0, intercept=2.0), name='A')
            comp.add_node(A)

            sched = Scheduler(**pytest.helpers.composition_to_scheduler_args(comp))
            sched.add_condition(A, AfterPass(0))

            termination_conds = {}
            termination_conds[TimeScale.ENVIRONMENT_SEQUENCE] = AfterNEnvironmentStateUpdates(1)
            termination_conds[TimeScale.ENVIRONMENT_STATE_UPDATE] = AtPass(5)
            output = list(sched.run(termination_conds=termination_conds))

            expected_output = [set(), A, A, A, A]
            assert output == pytest.helpers.setify_expected_output(expected_output)

        def test_AfterNPasses(self):
            comp = Composition()
            A = TransferMechanism(function=Linear(slope=5.0, intercept=2.0), name='A')
            comp.add_node(A)

            sched = Scheduler(**pytest.helpers.composition_to_scheduler_args(comp))
            sched.add_condition(A, AfterNPasses(1))

            termination_conds = {}
            termination_conds[TimeScale.ENVIRONMENT_SEQUENCE] = AfterNEnvironmentStateUpdates(1)
            termination_conds[TimeScale.ENVIRONMENT_STATE_UPDATE] = AtPass(5)
            output = list(sched.run(termination_conds=termination_conds))

            expected_output = [set(), A, A, A, A]
            assert output == pytest.helpers.setify_expected_output(expected_output)

        def test_BeforeEnvironmentStateUpdate(self):
            comp = Composition()
            A = TransferMechanism(function=Linear(slope=5.0, intercept=2.0), name='A')
            comp.add_node(A)

            sched = Scheduler(**pytest.helpers.composition_to_scheduler_args(comp))
            sched.add_condition(A, BeforeEnvironmentStateUpdate(4))

            termination_conds = {}
            termination_conds[TimeScale.ENVIRONMENT_SEQUENCE] = AfterNEnvironmentStateUpdates(5)
            termination_conds[TimeScale.ENVIRONMENT_STATE_UPDATE] = AtPass(1)
            comp.run(
                inputs={A: range(6)},
                scheduler=sched,
                termination_processing=termination_conds
            )
            output = sched.execution_list[comp.default_execution_id]

            expected_output = [A, A, A, A, set()]
            assert output == pytest.helpers.setify_expected_output(expected_output)

        def test_AtEnvironmentStateUpdate(self):
            comp = Composition()
            A = TransferMechanism(function=Linear(slope=5.0, intercept=2.0), name='A')
            comp.add_node(A)

            sched = Scheduler(**pytest.helpers.composition_to_scheduler_args(comp))
            sched.add_condition(A, Always())

            termination_conds = {}
            termination_conds[TimeScale.ENVIRONMENT_SEQUENCE] = AtEnvironmentStateUpdate(4)
            termination_conds[TimeScale.ENVIRONMENT_STATE_UPDATE] = AtPass(1)
            comp.run(
                inputs={A: range(6)},
                scheduler=sched,
                termination_processing=termination_conds
            )
            output = sched.execution_list[comp.default_execution_id]

            expected_output = [A, A, A, A]
            assert output == pytest.helpers.setify_expected_output(expected_output)

        def test_AfterEnvironmentStateUpdate(self):
            comp = Composition()
            A = TransferMechanism(function=Linear(slope=5.0, intercept=2.0), name='A')
            comp.add_node(A)

            sched = Scheduler(**pytest.helpers.composition_to_scheduler_args(comp))
            sched.add_condition(A, Always())

            termination_conds = {}
            termination_conds[TimeScale.ENVIRONMENT_SEQUENCE] = AfterEnvironmentStateUpdate(4)
            termination_conds[TimeScale.ENVIRONMENT_STATE_UPDATE] = AtPass(1)
            comp.run(
                inputs={A: range(6)},
                scheduler=sched,
                termination_processing=termination_conds
            )
            output = sched.execution_list[comp.default_execution_id]

            expected_output = [A, A, A, A, A]
            assert output == pytest.helpers.setify_expected_output(expected_output)

        def test_AfterNEnvironmentStateUpdates(self):
            comp = Composition()
            A = TransferMechanism(function=Linear(slope=5.0, intercept=2.0), name='A')
            comp.add_node(A)

            sched = Scheduler(**pytest.helpers.composition_to_scheduler_args(comp))
            sched.add_condition(A, AfterNPasses(1))

            termination_conds = {}
            termination_conds[TimeScale.ENVIRONMENT_SEQUENCE] = AfterNEnvironmentStateUpdates(1)
            termination_conds[TimeScale.ENVIRONMENT_STATE_UPDATE] = AtPass(5)
            output = list(sched.run(termination_conds=termination_conds))

            expected_output = [set(), A, A, A, A]
            assert output == pytest.helpers.setify_expected_output(expected_output)

    @pytest.mark.psyneulink
    class TestComponentBased:

        def test_BeforeNCalls(self):
            comp = Composition()
            A = TransferMechanism(function=Linear(slope=5.0, intercept=2.0), name='A')
            comp.add_node(A)

            sched = Scheduler(**pytest.helpers.composition_to_scheduler_args(comp))
            sched.add_condition(A, BeforeNCalls(A, 3))

            termination_conds = {}
            termination_conds[TimeScale.ENVIRONMENT_SEQUENCE] = AfterNEnvironmentStateUpdates(1)
            termination_conds[TimeScale.ENVIRONMENT_STATE_UPDATE] = AtPass(5)
            output = list(sched.run(termination_conds=termination_conds))

            expected_output = [A, A, A, set(), set()]
            assert output == pytest.helpers.setify_expected_output(expected_output)

        # NOTE:
        # The behavior is not desired (i.e. depending on the order mechanisms are checked, B running AtCall(A, x))
        # may run on both the xth and x+1st call of A; if A and B are not parent-child
        # A fix could invalidate key assumptions and affect many other conditions
        # Since this condition is unlikely to be used, it's best to leave it for now
        # def test_AtCall(self):
        #     comp = Composition()
        #     A = TransferMechanism(function = Linear(slope=5.0, intercept = 2.0), name = 'A')
        #     B = TransferMechanism(function = Linear(intercept = 4.0), name = 'B')
        #     C = TransferMechanism(function = Linear(intercept = 1.5), name = 'C')
        #     for m in [A,B]:
        #         comp.add_node(m)

        #     sched = Scheduler(**pytest.helpers.composition_to_scheduler_args(comp))
        #     sched.add_condition(A, Always())
        #     sched.add_condition(B, AtCall(A, 3))

        #     termination_conds = {}
        #     termination_conds[TimeScale.ENVIRONMENT_SEQUENCE] = AfterNEnvironmentStateUpdates(1)
        #     termination_conds[TimeScale.ENVIRONMENT_STATE_UPDATE] = AtPass(5)
        #     output = list(sched.run(termination_conds=termination_conds))

        #     expected_output = [A, A, set([A, B]), A, A]
        #     assert output == pytest.helpers.setify_expected_output(expected_output)

        def test_AfterCall(self):
            comp = Composition()
            A = TransferMechanism(function=Linear(slope=5.0, intercept=2.0), name='A')
            B = TransferMechanism(function=Linear(intercept=4.0), name='B')
            for m in [A, B]:
                comp.add_node(m)

            sched = Scheduler(**pytest.helpers.composition_to_scheduler_args(comp))
            sched.add_condition(B, AfterCall(A, 3))

            termination_conds = {}
            termination_conds[TimeScale.ENVIRONMENT_SEQUENCE] = AfterNEnvironmentStateUpdates(1)
            termination_conds[TimeScale.ENVIRONMENT_STATE_UPDATE] = AtPass(5)
            output = list(sched.run(termination_conds=termination_conds))

            expected_output = [A, A, A, set([A, B]), set([A, B])]
            assert output == pytest.helpers.setify_expected_output(expected_output)

        def test_AfterNCalls(self):
            comp = Composition()
            A = TransferMechanism(function=Linear(slope=5.0, intercept=2.0), name='A')
            B = TransferMechanism(function=Linear(intercept=4.0), name='B')
            for m in [A, B]:
                comp.add_node(m)

            sched = Scheduler(**pytest.helpers.composition_to_scheduler_args(comp))
            sched.add_condition(A, Always())
            sched.add_condition(B, AfterNCalls(A, 3))

            termination_conds = {}
            termination_conds[TimeScale.ENVIRONMENT_SEQUENCE] = AfterNEnvironmentStateUpdates(1)
            termination_conds[TimeScale.ENVIRONMENT_STATE_UPDATE] = AtPass(5)
            output = list(sched.run(termination_conds=termination_conds))

            expected_output = [A, A, set([A, B]), set([A, B]), set([A, B])]
            assert output == pytest.helpers.setify_expected_output(expected_output)

    @pytest.mark.psyneulink
    class TestConvenience:

        def test_AtEnvironmentStateUpdateStart(self):
            comp = Composition()
            A = TransferMechanism(name='A')
            B = TransferMechanism(name='B')
            comp.add_linear_processing_pathway([A, B])

            sched = Scheduler(**pytest.helpers.composition_to_scheduler_args(comp))
            sched.add_condition(B, AtEnvironmentStateUpdateStart())

            termination_conds = {
                TimeScale.ENVIRONMENT_STATE_UPDATE: AtPass(3)
            }
            output = list(sched.run(termination_conds=termination_conds))

            expected_output = [A, B, A, A]
            assert output == pytest.helpers.setify_expected_output(expected_output)

    @pytest.mark.psyneulink
    def test_composite_condition_multi(self):
        comp = Composition()
        A = TransferMechanism(function=Linear(slope=5.0, intercept=2.0), name='A')
        B = TransferMechanism(function=Linear(intercept=4.0), name='B')
        C = TransferMechanism(function=Linear(intercept=1.5), name='C')
        for m in [A, B, C]:
            comp.add_node(m)
        comp.add_projection(MappingProjection(), A, B)
        comp.add_projection(MappingProjection(), B, C)
        sched = Scheduler(**pytest.helpers.composition_to_scheduler_args(comp))

        sched.add_condition(A, EveryNPasses(1))
        sched.add_condition(B, EveryNCalls(A, 2))
        sched.add_condition(C, All(
            Any(
                AfterPass(6),
                AfterNCalls(B, 2)
            ),
            Any(
                AfterPass(2),
                AfterNCalls(B, 3)
            )
        )
        )

        termination_conds = {}
        termination_conds[TimeScale.ENVIRONMENT_SEQUENCE] = AfterNEnvironmentStateUpdates(1)
        termination_conds[TimeScale.ENVIRONMENT_STATE_UPDATE] = AfterNCalls(C, 3)
        output = list(sched.run(termination_conds=termination_conds))
        expected_output = [
            A, A, B, A, A, B, C, A, C, A, B, C
        ]
        assert output == pytest.helpers.setify_expected_output(expected_output)

    @pytest.mark.psyneulink
    def test_AfterNCallsCombined(self):
        comp = Composition()
        A = TransferMechanism(function=Linear(slope=5.0, intercept=2.0), name='A')
        A.is_finished_flag = True
        B = TransferMechanism(function=Linear(intercept=4.0), name='B')
        B.is_finished_flag = True
        C = TransferMechanism(function=Linear(intercept=1.5), name='C')
        for m in [A, B, C]:
            comp.add_node(m)
        comp.add_projection(MappingProjection(), A, B)
        comp.add_projection(MappingProjection(), B, C)
        sched = Scheduler(**pytest.helpers.composition_to_scheduler_args(comp))

        sched.add_condition(A, EveryNPasses(1))
        sched.add_condition(B, EveryNCalls(A, 2))
        sched.add_condition(C, EveryNCalls(B, 2))

        termination_conds = {}
        termination_conds[TimeScale.ENVIRONMENT_SEQUENCE] = AfterNEnvironmentStateUpdates(1)
        termination_conds[TimeScale.ENVIRONMENT_STATE_UPDATE] = AfterNCallsCombined(B, C, n=4)
        output = list(sched.run(termination_conds=termination_conds))

        expected_output = [
            A, A, B, A, A, B, C, A, A, B
        ]
        assert output == pytest.helpers.setify_expected_output(expected_output)

    @pytest.mark.psyneulink
    def test_AllHaveRun(self):
        comp = Composition()
        A = TransferMechanism(function=Linear(slope=5.0, intercept=2.0), name='A')
        A.is_finished_flag = False
        B = TransferMechanism(function=Linear(intercept=4.0), name='B')
        B.is_finished_flag = True
        C = TransferMechanism(function=Linear(intercept=1.5), name='C')
        for m in [A, B, C]:
            comp.add_node(m)
        comp.add_projection(MappingProjection(), A, B)
        comp.add_projection(MappingProjection(), B, C)
        sched = Scheduler(**pytest.helpers.composition_to_scheduler_args(comp))

        sched.add_condition(A, EveryNPasses(1))
        sched.add_condition(B, EveryNCalls(A, 2))
        sched.add_condition(C, EveryNCalls(B, 2))

        termination_conds = {}
        termination_conds[TimeScale.ENVIRONMENT_SEQUENCE] = AfterNEnvironmentStateUpdates(1)
        termination_conds[TimeScale.ENVIRONMENT_STATE_UPDATE] = AllHaveRun(A, B, C)
        output = list(sched.run(termination_conds=termination_conds))

        expected_output = [
            A, A, B, A, A, B, C
        ]
        assert output == pytest.helpers.setify_expected_output(expected_output)

    @pytest.mark.psyneulink
    def test_AllHaveRun_2(self):
        comp = Composition()
        A = TransferMechanism(function=Linear(slope=5.0, intercept=2.0), name='A')
        B = TransferMechanism(function=Linear(intercept=4.0), name='B')
        C = TransferMechanism(function=Linear(intercept=1.5), name='C')
        for m in [A, B, C]:
            comp.add_node(m)
        comp.add_projection(MappingProjection(), A, B)
        comp.add_projection(MappingProjection(), B, C)
        sched = Scheduler(**pytest.helpers.composition_to_scheduler_args(comp))

        sched.add_condition(A, EveryNPasses(1))
        sched.add_condition(B, EveryNCalls(A, 2))
        sched.add_condition(C, EveryNCalls(B, 2))

        termination_conds = {}
        termination_conds[TimeScale.ENVIRONMENT_SEQUENCE] = AfterNEnvironmentStateUpdates(1)
        termination_conds[TimeScale.ENVIRONMENT_STATE_UPDATE] = AllHaveRun()
        output = list(sched.run(termination_conds=termination_conds))

        expected_output = [
            A, A, B, A, A, B, C
        ]
        assert output == pytest.helpers.setify_expected_output(expected_output)


@pytest.mark.psyneulink
class TestWhenFinished:

    @classmethod
    def setup_class(self):
        self.orig_is_finished_flag = TransferMechanism.is_finished_flag
        self.orig_is_finished = TransferMechanism.is_finished
        TransferMechanism.is_finished_flag = True
        TransferMechanism.is_finished = lambda self, context: self.is_finished_flag

    @classmethod
    def teardown_class(self):
        del TransferMechanism.is_finished_flag
        del TransferMechanism.is_finished
        TransferMechanism.is_finished_flag = self.orig_is_finished_flag
        TransferMechanism.is_finished = self.orig_is_finished

    def test_WhenFinishedAny_1(self):
        comp = Composition()
        A = TransferMechanism(function=Linear(slope=5.0, intercept=2.0), name='A')
        A.is_finished_flag = True
        B = TransferMechanism(function=Linear(intercept=4.0), name='B')
        B.is_finished_flag = True
        C = TransferMechanism(function=Linear(intercept=1.5), name='C')
        for m in [A, B, C]:
            comp.add_node(m)
        comp.add_projection(MappingProjection(), A, C)
        comp.add_projection(MappingProjection(), B, C)
        sched = Scheduler(**pytest.helpers.composition_to_scheduler_args(comp))

        sched.add_condition(A, EveryNPasses(1))
        sched.add_condition(B, EveryNPasses(1))
        sched.add_condition(C, WhenFinishedAny(A, B))

        termination_conds = {}
        termination_conds[TimeScale.ENVIRONMENT_SEQUENCE] = AfterNEnvironmentStateUpdates(1)
        termination_conds[TimeScale.ENVIRONMENT_STATE_UPDATE] = AfterNCalls(C, 1)
        output = list(sched.run(termination_conds=termination_conds))
        expected_output = [
            set([A, B]), C
        ]
        assert output == pytest.helpers.setify_expected_output(expected_output)

    def test_WhenFinishedAny_2(self):
        comp = Composition()
        A = TransferMechanism(function=Linear(slope=5.0, intercept=2.0), name='A')
        A.is_finished_flag = False
        B = TransferMechanism(function=Linear(intercept=4.0), name='B')
        B.is_finished_flag = True
        C = TransferMechanism(function=Linear(intercept=1.5), name='C')
        for m in [A, B, C]:
            comp.add_node(m)
        comp.add_projection(MappingProjection(), A, C)
        comp.add_projection(MappingProjection(), B, C)
        sched = Scheduler(**pytest.helpers.composition_to_scheduler_args(comp))

        sched.add_condition(A, EveryNPasses(1))
        sched.add_condition(B, EveryNPasses(1))
        sched.add_condition(C, WhenFinishedAny(A, B))

        termination_conds = {}
        termination_conds[TimeScale.ENVIRONMENT_SEQUENCE] = AfterNEnvironmentStateUpdates(1)
        termination_conds[TimeScale.ENVIRONMENT_STATE_UPDATE] = AfterNCalls(A, 5)
        output = list(sched.run(termination_conds=termination_conds))
        expected_output = [
            set([A, B]), C, set([A, B]), C, set([A, B]), C, set([A, B]), C, set([A, B])
        ]
        assert output == pytest.helpers.setify_expected_output(expected_output)

    def test_WhenFinishedAny_noargs(self):
        comp = Composition()
        A = TransferMechanism(function=Linear(slope=5.0, intercept=2.0), name='A')
        B = TransferMechanism(function=Linear(intercept=4.0), name='B')
        C = TransferMechanism(function=Linear(intercept=1.5), name='C')
        for m in [A, B, C]:
            m.is_finished_flag = False
            comp.add_node(m)
        comp.add_projection(MappingProjection(), A, C)
        comp.add_projection(MappingProjection(), B, C)
        sched = Scheduler(**pytest.helpers.composition_to_scheduler_args(comp))

        sched.add_condition(A, Always())
        sched.add_condition(B, Always())
        sched.add_condition(C, Always())

        termination_conds = {}
        termination_conds[TimeScale.ENVIRONMENT_SEQUENCE] = AfterNEnvironmentStateUpdates(1)
        termination_conds[TimeScale.ENVIRONMENT_STATE_UPDATE] = WhenFinishedAny()
        output = []
        i = 0
        for step in sched.run(termination_conds=termination_conds):
            if i == 3:
                A.is_finished_flag = True
                B.is_finished_flag = True
            if i == 4:
                C.is_finished_flag = True
            output.append(step)
            i += 1
        expected_output = [
            set([A, B]), C, set([A, B]), C,
        ]
        assert output == pytest.helpers.setify_expected_output(expected_output)

    def test_WhenFinishedAll_1(self):
        comp = Composition()
        A = TransferMechanism(function=Linear(slope=5.0, intercept=2.0), name='A')
        A.is_finished_flag = True
        B = TransferMechanism(function=Linear(intercept=4.0), name='B')
        B.is_finished_flag = True
        C = TransferMechanism(function=Linear(intercept=1.5), name='C')
        for m in [A, B, C]:
            comp.add_node(m)
        comp.add_projection(MappingProjection(), A, C)
        comp.add_projection(MappingProjection(), B, C)
        sched = Scheduler(**pytest.helpers.composition_to_scheduler_args(comp))

        sched.add_condition(A, EveryNPasses(1))
        sched.add_condition(B, EveryNPasses(1))
        sched.add_condition(C, WhenFinishedAll(A, B))

        termination_conds = {}
        termination_conds[TimeScale.ENVIRONMENT_SEQUENCE] = AfterNEnvironmentStateUpdates(1)
        termination_conds[TimeScale.ENVIRONMENT_STATE_UPDATE] = AfterNCalls(C, 1)
        output = list(sched.run(termination_conds=termination_conds))
        expected_output = [
            set([A, B]), C
        ]
        assert output == pytest.helpers.setify_expected_output(expected_output)

    def test_WhenFinishedAll_2(self):
        comp = Composition()
        A = TransferMechanism(function=Linear(slope=5.0, intercept=2.0), name='A')
        A.is_finished_flag = False
        B = TransferMechanism(function=Linear(intercept=4.0), name='B')
        B.is_finished_flag = True
        C = TransferMechanism(function=Linear(intercept=1.5), name='C')
        for m in [A, B, C]:
            comp.add_node(m)
        comp.add_projection(MappingProjection(), A, C)
        comp.add_projection(MappingProjection(), B, C)
        sched = Scheduler(**pytest.helpers.composition_to_scheduler_args(comp))

        sched.add_condition(A, EveryNPasses(1))
        sched.add_condition(B, EveryNPasses(1))
        sched.add_condition(C, WhenFinishedAll(A, B))

        termination_conds = {}
        termination_conds[TimeScale.ENVIRONMENT_SEQUENCE] = AfterNEnvironmentStateUpdates(1)
        termination_conds[TimeScale.ENVIRONMENT_STATE_UPDATE] = AfterNCalls(A, 5)
        output = list(sched.run(termination_conds=termination_conds))
        expected_output = [
            set([A, B]), set([A, B]), set([A, B]), set([A, B]), set([A, B]),
        ]
        assert output == pytest.helpers.setify_expected_output(expected_output)

    def test_WhenFinishedAll_noargs(self):
        comp = Composition()
        A = TransferMechanism(function=Linear(slope=5.0, intercept=2.0), name='A')
        B = TransferMechanism(function=Linear(intercept=4.0), name='B')
        C = TransferMechanism(function=Linear(intercept=1.5), name='C')
        for m in [A, B, C]:
            comp.add_node(m)
            m.is_finished_flag = False
        comp.add_projection(MappingProjection(), A, C)
        comp.add_projection(MappingProjection(), B, C)
        sched = Scheduler(**pytest.helpers.composition_to_scheduler_args(comp))

        sched.add_condition(A, Always())
        sched.add_condition(B, Always())
        sched.add_condition(C, Always())

        termination_conds = {}
        termination_conds[TimeScale.ENVIRONMENT_SEQUENCE] = AfterNEnvironmentStateUpdates(1)
        termination_conds[TimeScale.ENVIRONMENT_STATE_UPDATE] = WhenFinishedAll()
        output = []
        i = 0
        for step in sched.run(termination_conds=termination_conds):
            if i == 3:
                A.is_finished_flag = True
                B.is_finished_flag = True
            if i == 4:
                C.is_finished_flag = True
            output.append(step)
            i += 1
        expected_output = [
            set([A, B]), C, set([A, B]), C, set([A, B]),
        ]
        assert output == pytest.helpers.setify_expected_output(expected_output)


@pytest.mark.psyneulink
class TestAbsolute:
    A = TransferMechanism(name='scheduler-pytests-A')
    B = TransferMechanism(name='scheduler-pytests-B')
    C = TransferMechanism(name='scheduler-pytests-C')

    @pytest.mark.parametrize(
        'conditions, termination_conds',
        [
            (
                {A: TimeInterval(repeat=8), B: TimeInterval(repeat=4), C: TimeInterval(repeat=2)},
                {TimeScale.ENVIRONMENT_STATE_UPDATE: AfterNCalls(A, 2)},
            ),
            (
                {A: TimeInterval(repeat=5), B: TimeInterval(repeat=3), C: TimeInterval(repeat=1)},
                {TimeScale.ENVIRONMENT_STATE_UPDATE: AfterNCalls(A, 2)},
            ),
            (
                {A: TimeInterval(repeat=3), B: TimeInterval(repeat=2)},
                {TimeScale.ENVIRONMENT_STATE_UPDATE: AfterNCalls(A, 2)},
            ),
            (
                {A: TimeInterval(repeat=5), B: TimeInterval(repeat=7)},
                {TimeScale.ENVIRONMENT_STATE_UPDATE: AfterNCalls(B, 2)},
            ),
            (
                {A: TimeInterval(repeat=1200), B: TimeInterval(repeat=1000)},
                {TimeScale.ENVIRONMENT_STATE_UPDATE: AfterNCalls(A, 3)},
            ),
            (
                {A: TimeInterval(repeat=0.33333), B: TimeInterval(repeat=0.66666)},
                {TimeScale.ENVIRONMENT_STATE_UPDATE: AfterNCalls(B, 3)},
            ),
            # smaller than default units cause floating point issue without mitigation
            (
                {A: TimeInterval(repeat=2 * _unit_registry.us), B: TimeInterval(repeat=4 * _unit_registry.us)},
                {TimeScale.ENVIRONMENT_STATE_UPDATE: AfterNCalls(B, 3)},
            ),
        ]
    )
    def test_TimeInterval_linear_everynms(self, conditions, termination_conds):
        comp = Composition()

        comp.add_linear_processing_pathway([self.A, self.B, self.C])
        comp.scheduler.add_condition_set(conditions)

        list(comp.scheduler.run(termination_conds=termination_conds))

        for node, cond in conditions.items():
            executions = [
                comp.scheduler.execution_timestamps[comp.default_execution_id][i].absolute
                for i in range(len(comp.scheduler.execution_list[comp.default_execution_id]))
                if node in comp.scheduler.execution_list[comp.default_execution_id][i]
            ]

            for i in range(1, len(executions)):
                assert (executions[i] - executions[i - 1]) == cond.repeat

    @pytest.mark.parametrize(
        'conditions, termination_conds',
        [
            (
                {
                    A: TimeInterval(repeat=10, start=100),
                    B: TimeInterval(repeat=10, start=300),
                    C: TimeInterval(repeat=10, start=400)
                },
                {TimeScale.ENVIRONMENT_STATE_UPDATE: TimeInterval(start=500)}
            ),
            (
                {
                    A: TimeInterval(start=100),
                    B: TimeInterval(start=300),
                    C: TimeInterval(start=400)
                },
                {TimeScale.ENVIRONMENT_STATE_UPDATE: TimeInterval(start=500)}
            ),
            (
                {
                    A: TimeInterval(repeat=2, start=105),
                    B: TimeInterval(repeat=7, start=317),
                    C: TimeInterval(repeat=11, start=431)
                },
                {TimeScale.ENVIRONMENT_STATE_UPDATE: TimeInterval(start=597)}
            ),
        ]
    )
    def test_TimeInterval_no_dependencies(self, conditions, termination_conds):
        comp = Composition()
        comp.add_nodes([self.A, self.B, self.C])
        comp.scheduler.add_condition_set(conditions)
        consideration_set_execution_abs_value = comp.scheduler._get_absolute_consideration_set_execution_unit(termination_conds)

        list(comp.scheduler.run(termination_conds=termination_conds))

        for node, cond in conditions.items():
            executions = [
                comp.scheduler.execution_timestamps[comp.default_execution_id][i].absolute
                for i in range(len(comp.scheduler.execution_list[comp.default_execution_id]))
                if node in comp.scheduler.execution_list[comp.default_execution_id][i]
            ]

            for i in range(1, len(executions)):
                interval = (executions[i] - executions[i - 1])

                if cond.repeat is not None:
                    assert interval == cond.repeat
                else:
                    assert interval == consideration_set_execution_abs_value

            if cond.start is not None:
                if cond.start_inclusive:
                    assert cond.start in executions
                else:
                    assert cond.start + consideration_set_execution_abs_value in executions

    @pytest.mark.parametrize(
        'repeat, unit, expected_repeat',
        [
            (1, None, 1 * _unit_registry.ms),
            ('1ms', None, 1 * _unit_registry.ms),
            (1 * _unit_registry.ms, None, 1 * _unit_registry.ms),
            (1, 'ms', 1 * _unit_registry.ms),
            (1, _unit_registry.ms, 1 * _unit_registry.ms),
            ('1', _unit_registry.ms, 1 * _unit_registry.ms),
            (1 * _unit_registry.ms, _unit_registry.ns, 1 * _unit_registry.ms),
            (1000 * _unit_registry.ms, None, 1000 * _unit_registry.ms),
        ]
    )
    def test_TimeInterval_time_specs(self, repeat, unit, expected_repeat):
        if unit is None:
            c = TimeInterval(repeat=repeat)
        else:
            c = TimeInterval(repeat=repeat, unit=unit)

        assert c.repeat == expected_repeat

    @pytest.mark.parametrize(
        'repeat, inclusive, last_time',
        [
            (10, True, 10 * _unit_registry.ms),
            (10, False, 11 * _unit_registry.ms),
        ]
    )
    def test_TimeTermination(
        self,
        three_node_linear_composition,
        repeat,
        inclusive,
        last_time
    ):
        _, comp = three_node_linear_composition

        comp.scheduler.termination_conds = {
            TimeScale.ENVIRONMENT_STATE_UPDATE: TimeTermination(repeat, inclusive)
        }
        list(comp.scheduler.run())

        assert comp.scheduler.get_clock(comp.scheduler.default_execution_id).time.absolute == last_time



"""
This module provides utilities used to schedule the execution of psyneulink components

https://princetonuniversity.github.io/PsyNeuLink/Scheduling.html
"""

import pint

_unit_registry = pint.get_application_registry()
pint.set_application_registry(_unit_registry)
_unit_registry.precision = 8  # TODO: remove when floating point issues resolved

from . import condition  # noqa: E402
from . import scheduler  # noqa: E402
from . import time  # noqa: E402

from .condition import *  # noqa: E402
from .scheduler import *  # noqa: E402
from .time import *  # noqa: E402

__all__ = list(condition.__all__)
__all__.extend(scheduler.__all__)
__all__.extend(time.__all__)
__all__.extend([
    '_unit_registry'
])

from . import _version  # noqa: E402
__version__ = _version.get_versions()['version']

del pint



import types

import psyneulink as pnl
import pytest


def pytest_assertrepr_compare(op, left, right):
    if isinstance(left, list) and isinstance(right, list) and op == '==':
        return [
            'Time Step output matching:',
            'Actual output:', str(left),
            'Expected output:', str(right)
        ]


@pytest.helpers.register
def setify_expected_output(expected_output):
    type_set = type(set())
    for i in range(len(expected_output)):
        if type(expected_output[i]) is not type_set:
            try:
                iter(expected_output[i])
                expected_output[i] = set(expected_output[i])
            except TypeError:
                expected_output[i] = set([expected_output[i]])
    return expected_output


@pytest.helpers.register
def create_graph_from_pathways(*pathways):
    dependency_dict = {}
    for p in pathways:
        for i in range(len(p)):
            if p[i] not in dependency_dict:
                dependency_dict[p[i]] = set()

            try:
                dependency_dict[p[i + 1]].add(p[i])
            except KeyError:
                dependency_dict[p[i + 1]] = {p[i]}
            except IndexError:
                pass

    return dependency_dict


@pytest.helpers.register
def create_node(function=lambda x: x):
    return types.SimpleNamespace(function=function)


@pytest.fixture
def three_node_linear_composition():
    A = pnl.TransferMechanism(name='A')
    B = pnl.TransferMechanism(name='B')
    C = pnl.TransferMechanism(name='C')

    comp = pnl.Composition()
    comp.add_linear_processing_pathway([A, B, C])

    return comp.nodes, comp
